@@ -1,6 +1,6 @@
 /* _NVRM_COPYRIGHT_BEGIN_
  *
- * Copyright 2001-2011 by NVIDIA Corporation.  All rights reserved.  All
+ * Copyright 2001-2013 by NVIDIA Corporation.  All rights reserved.  All
  * information contained herein is proprietary and confidential to NVIDIA
  * Corporation.  Any use, reproduction, or disclosure without the written
  * permission of NVIDIA Corporation is prohibited.
@@ -32,9 +32,9 @@ RM_STATUS NV_API_CALL os_alloc_mem(
     if (alloc_size != size)
         return RM_ERR_INVALID_LIMIT;
 
-    /* XXX Fix me? (malloc flags) */
+    // XXX Fix me? (malloc flags)
     *address = malloc(alloc_size, M_NVIDIA, M_NOWAIT | M_ZERO);
-    return *address ? RM_OK : RM_ERR_NO_FREE_MEM;
+    return *address ? RM_OK : RM_ERR_NO_MEMORY;
 }
 
 void NV_API_CALL os_free_mem(void *address)
@@ -193,7 +193,8 @@ void NV_API_CALL os_io_write_dword(
 void* NV_API_CALL os_map_kernel_space(
     NvU64 start,
     NvU64 size,
-    NvU32 mode
+    NvU32 mode,
+    NvU32 memType
 )
 {
     int map_mode;
@@ -250,49 +251,6 @@ void NV_API_CALL os_unmap_user_space(
 {
 }
 
-RM_STATUS NV_API_CALL os_set_mem_range(
-    NvU64 start,
-    NvU64 size,
-    NvU32 mode,
-    NvS32 *memory_token
-)
-{
-    int arg;
-    struct mem_range_desc mrd;
-
-    mrd.mr_base  = start;
-    mrd.mr_len   = size;
-    mrd.mr_flags = MDF_WRITECOMBINE;
-
-    strcpy(mrd.mr_owner, "NVIDIA");
-    arg = MEMRANGE_SET_UPDATE;
-
-    if (mem_range_attr_set(&mrd, &arg))
-        return RM_ERROR;
-
-    return RM_OK;
-}
-
-RM_STATUS NV_API_CALL os_unset_mem_range(
-    NvU64 start,
-    NvU64 size,
-    NvS32 memory_token
-)
-{
-    int arg;
-    struct mem_range_desc mrd;
-
-    mrd.mr_base = start;
-    mrd.mr_len  = size;
-    arg = MEMRANGE_SET_REMOVE;
-
-    if (mem_range_attr_set(&mrd, &arg))
-        return RM_ERROR;
-
-    return RM_OK;
-}
-
-
 /*
  * The current debug level is used to determine if certain debug messages
  * are printed to the system console/log files or not. It defaults to the
@@ -445,18 +403,38 @@ NvS32 NV_API_CALL os_mem_cmp(
 }
 
 NvU8* NV_API_CALL os_mem_copy(
-    NvU8 *dst,
+    NvU8       *dst,
     const NvU8 *src,
-    NvU32 length
+    NvU32       length
 )
 {
-#if defined(NVCPU_X86_64)
-    uint32_t i;
-    for (i = 0; i < length; i++) dst[i] = src[i];
-    return dst;
-#else
-    return memcpy(dst, src, length);
-#endif
+    NvU8 *ret = dst;
+    NvU32 dwords, bytes = length;
+
+    if ((length >= 128) &&
+        (((NvUPtr)dst & 3) == 0) & (((NvUPtr)src & 3) == 0))
+    {
+        dwords = (length / sizeof(NvU32));
+        bytes = (length % sizeof(NvU32));
+
+        while (dwords != 0)
+        {
+            *(NvU32 *)dst = *(const NvU32 *)src;
+            dst += sizeof(NvU32);
+            src += sizeof(NvU32);
+            dwords--;
+        }
+    }
+
+    while (bytes != 0)
+    {
+        *dst = *src;
+        dst++;
+        src++;
+        bytes--;
+    }
+
+    return ret;
 }
 
 RM_STATUS NV_API_CALL os_memcpy_from_user(
@@ -478,12 +456,22 @@ RM_STATUS NV_API_CALL os_memcpy_to_user(
 }
 
 void* NV_API_CALL os_mem_set(
-    void *b,
-    NvU8  c,
-    NvU32 length
+    void  *dst,
+    NvU8   c,
+    NvU32  length
 )
 {
-    return memset(b, c, length);
+    NvU8 *ret = dst;
+    NvU32 bytes = length;
+
+    while (bytes != 0)
+    {
+        *(NvU8 *)dst = c;
+        dst = ((NvU8 *)dst + 1);
+        bytes--;
+    }
+
+    return ret;
 }
 
 char* NV_API_CALL os_string_copy(
@@ -499,15 +487,6 @@ NvU32 NV_API_CALL os_string_length(const char* s)
     return strlen(s);
 }
 
-RM_STATUS NV_API_CALL os_strncpy_from_user(
-    char *dst,
-    const char *src,
-    NvU32 n
-)
-{
-    return copyinstr(src, dst, n, NULL) ? RM_ERR_INVALID_POINTER : RM_OK;
-}
-
 NvU32 NV_API_CALL os_get_page_size(void)
 {
     return PAGE_SIZE;
@@ -518,9 +497,14 @@ NvU64 NV_API_CALL os_get_page_mask(void)
     return ~PAGE_MASK;
 }
 
-NvU64 NV_API_CALL os_get_system_memory_size(void)
+NvU8 NV_API_CALL os_get_page_shift(void)
 {
-    return ((NvU64)physmem * PAGE_SIZE) / RM_PAGE_SIZE;
+    return PAGE_SHIFT;
+}
+
+NvU64 NV_API_CALL os_get_num_phys_pages(void)
+{
+    return (NvU64)physmem;
 }
 
 NvU32 NV_API_CALL os_get_cpu_count(void)
@@ -528,6 +512,11 @@ NvU32 NV_API_CALL os_get_cpu_count(void)
     return mp_ncpus;
 }
 
+NvU32 NV_API_CALL os_get_cpu_number(void)
+{
+    return curcpu;
+}
+
 #ifdef NV_USE_WBINVD
 static void wbinvd_action_func(void *arg)
 {
@@ -559,6 +548,11 @@ RM_STATUS NV_API_CALL os_flush_user_cache(
     return RM_ERR_NOT_SUPPORTED;
 }
 
+RM_STATUS NV_API_CALL os_flush_cpu_cache_all(void)
+{
+    return RM_ERR_NOT_SUPPORTED;
+}
+
 static void sfence_action_func(void *arg)
 {
     __asm__ __volatile__("sfence": : :"memory");
@@ -580,116 +574,6 @@ RM_STATUS NV_API_CALL os_clear_smp_barrier(void)
     return RM_ERR_NOT_SUPPORTED;
 }
 
-void NV_API_CALL os_disable_preemption(void)
-{
-}
-
-void NV_API_CALL os_enable_preemption(void)
-{
-}
-
-struct os_mutex {
-    nv_stack_t *sp;
-    struct mtx mutex_mtx;
-    struct cv mutex_cv;
-    int refcnt;
-};
-
-RM_STATUS NV_API_CALL os_alloc_sema(void **semaphore)
-{
-    nv_stack_t *sp;
-    struct os_mutex *mtx;
-    RM_STATUS status;
-
-    NV_UMA_ZONE_ALLOC_STACK(sp);
-    if (sp == NULL)
-        return RM_ERR_NO_FREE_MEM;
-
-    status = os_alloc_mem((void **)&mtx, sizeof(struct os_mutex));
-    if (status != RM_OK) {
-        NV_UMA_ZONE_FREE_STACK(sp);
-        return status;
-    }
-
-    mtx_init(&mtx->mutex_mtx, "os.mutex_mtx", NULL, MTX_DEF | MTX_RECURSE);
-    cv_init(&mtx->mutex_cv, "os.mutex_cv");
-
-    mtx->sp = sp;
-    mtx->refcnt = 1;
-
-    *semaphore = (void *)mtx;
-
-    return RM_OK;
-}
-
-void NV_API_CALL os_free_sema(void *semaphore)
-{
-    struct os_mutex *mtx = semaphore;
-    nv_stack_t *sp;
-
-    sp = mtx->sp;
-    NV_UMA_ZONE_FREE_STACK(sp);
-
-    mtx_destroy(&mtx->mutex_mtx);
-    cv_destroy(&mtx->mutex_cv);
-
-    os_free_mem(semaphore);
-}
-
-RM_STATUS NV_API_CALL os_acquire_sema(void *semaphore)
-{
-    struct os_mutex *mtx = semaphore;
-
-    mtx_lock(&mtx->mutex_mtx);
-    if (mtx->refcnt > 0)
-        rm_disable_interrupts(mtx->sp);
-    mtx->refcnt--;
-    if (mtx->refcnt < 0)
-        cv_wait(&mtx->mutex_cv, &mtx->mutex_mtx);
-    mtx_unlock(&mtx->mutex_mtx);
-
-    return RM_OK;
-}
-
-RM_STATUS NV_API_CALL os_cond_acquire_sema(void *semaphore)
-{
-    struct os_mutex *mtx = semaphore;
-
-    mtx_lock(&mtx->mutex_mtx);
-    if (mtx->refcnt < 1) {
-        mtx_unlock(&mtx->mutex_mtx);
-        return RM_ERROR;
-    } else {
-        rm_disable_interrupts(mtx->sp);
-        mtx->refcnt--;
-        mtx_unlock(&mtx->mutex_mtx);
-    }
-
-    return RM_OK;
-}
-
-RM_STATUS NV_API_CALL os_release_sema(void *semaphore)
-{
-    struct os_mutex *mtx = semaphore;
-
-    mtx_lock(&mtx->mutex_mtx);
-    if (mtx->refcnt < 0)
-        cv_signal(&mtx->mutex_cv);
-    else
-        rm_enable_interrupts(mtx->sp);
-    mtx->refcnt++;
-    mtx_unlock(&mtx->mutex_mtx);
-
-    return RM_OK;
-}
-
-BOOL NV_API_CALL os_is_acquired_sema(void *semaphore)
-{
-    struct os_mutex *mtx = semaphore;
-    return (mtx->refcnt < 1);
-}
-
-
 RM_STATUS NV_API_CALL os_alloc_mutex(void **mutex)
 {
     RM_STATUS status;
@@ -832,6 +716,15 @@ BOOL NV_API_CALL os_is_efi_enabled(void)
     return FALSE;
 }
 
+RM_STATUS NV_API_CALL os_get_efi_screen_info(
+    NvU64 *pPhysicalAddress,
+    NvU32 *pHeight,
+    NvU32 *pPitch
+)
+{
+    return RM_ERR_NOT_SUPPORTED;
+}
+
 void NV_API_CALL os_register_compatible_ioctl(
     NvU32 cmd,
     NvU32 size
@@ -923,3 +816,60 @@ RM_STATUS NV_API_CALL os_get_version_info(os_version_info * pOsVersionInfo)
 {
     return RM_ERR_NOT_SUPPORTED;
 }
+
+NvBool NV_API_CALL os_is_xen_dom0(void)
+{
+    return FALSE;
+}
+
+NvBool NV_API_CALL os_is_vgx_hyper(void)
+{
+    return FALSE;
+}
+
+void NV_API_CALL os_bug_check(NvU32 bugCode, const char *bugCodeStr)
+{
+}
+
+RM_STATUS NV_API_CALL os_lock_user_pages(
+    void   *address,
+    NvU64   page_count,
+    void  **page_array
+)
+{
+    return RM_ERR_NOT_SUPPORTED;
+}
+
+RM_STATUS NV_API_CALL os_unlock_user_pages(
+    NvU64  page_count,
+    void  *page_array
+)
+{
+    return RM_ERR_NOT_SUPPORTED;
+}
+
+RM_STATUS NV_API_CALL os_match_mmap_offset(
+    void  *pAllocPrivate,
+    NvU64  offset,
+    NvU64 *pPageIndex
+)
+{
+    struct nvidia_alloc *at = pAllocPrivate;
+    NvU64 i;
+
+    for (i = 0; i < (at->size / PAGE_SIZE); i++) {
+        if (at->alloc_type_contiguous) {
+            if (offset == (at->pte_array[0].physical_address + (i * PAGE_SIZE))) {
+                *pPageIndex = i;
+                return RM_OK;
+            }
+        } else {
+            if (offset == at->pte_array[i].physical_address) {
+                *pPageIndex = i;
+                return RM_OK;
+            }
+        }
+    }
+
+    return RM_ERR_OBJECT_NOT_FOUND;
+}
