@@ -33,18 +33,18 @@ int nvidia_attach(device_t dev)
     sc = device_get_softc(dev);
     nv = sc->nv_state;
 
-    nv->os_state         = sc;
-    nv->flags            = 0;
-    nv->domain           = pci_get_domain(dev);
-    nv->bus              = pci_get_bus(dev);
-    nv->slot             = pci_get_slot(dev);
-    nv->vendor_id        = pci_get_vendor(dev);
-    nv->device_id        = pci_get_device(dev);
-    nv->handle           = dev;
+    nv->os_state           = sc;
+    nv->flags              = 0;
+    nv->pci_info.domain    = pci_get_domain(dev);
+    nv->pci_info.bus       = pci_get_bus(dev);
+    nv->pci_info.slot      = pci_get_slot(dev);
+    nv->pci_info.vendor_id = pci_get_vendor(dev);
+    nv->pci_info.device_id = pci_get_device(dev);
+    nv->handle             = dev;
 
     for (i = 0; i < NV_GPU_NUM_BARS; i++) {
         if (sc->BAR_recs[i] != NULL) {
-            nv->bars[i].address = rman_get_start(sc->BAR_recs[i]);
+            nv->bars[i].cpu_address = rman_get_start(sc->BAR_recs[i]);
             nv->bars[i].strapped_size = rman_get_size(sc->BAR_recs[i]);
             nv->bars[i].size = nv->bars[i].strapped_size;
         }
@@ -68,7 +68,8 @@ int nvidia_attach(device_t dev)
                 "NVRM: Supported NVIDIA GPU Products' in this release's README,\n"
                 "NVRM: available on the FreeBSD graphics driver download page at\n"
                 "NVRM: www.nvidia.com.\n",
-                nv->bus, nv->slot, nv->vendor_id, nv->device_id, NV_VERSION_STRING);
+                nv->pci_info.bus, nv->pci_info.slot, nv->pci_info.vendor_id, 
+                nv->pci_info.device_id, NV_VERSION_STRING);
         }
         return ENXIO;
     }
@@ -80,7 +81,7 @@ int nvidia_attach(device_t dev)
                 goto failed;
 
             sglist_append_phys(sc->BAR_sg_lists[i],
-                    nv->bars[i].address, nv->bars[i].size);
+                    nv->bars[i].cpu_address, nv->bars[i].size);
 
             sc->BAR_objects[i] = NV_VM_PAGER_ALLOCATE(OBJT_SG,
                     sc->BAR_sg_lists[i],
@@ -89,7 +90,7 @@ int nvidia_attach(device_t dev)
             if (!sc->BAR_objects[i])
                 goto failed;
 
-            VM_OBJECT_LOCK(sc->BAR_objects[i]);
+            VM_OBJECT_WLOCK(sc->BAR_objects[i]);
             switch (i) {
                 case NV_GPU_BAR_INDEX_FB:
                     vm_object_set_memattr(sc->BAR_objects[i],
@@ -101,7 +102,7 @@ int nvidia_attach(device_t dev)
                             VM_MEMATTR_UNCACHEABLE);
                     break;
             }
-            VM_OBJECT_UNLOCK(sc->BAR_objects[i]);
+            VM_OBJECT_WUNLOCK(sc->BAR_objects[i]);
         }
     }
 
@@ -190,6 +191,8 @@ int nvidia_suspend(device_t dev)
     NV_PCI_CHECK_CONFIG_SPACE(sp, nv, TRUE, TRUE, TRUE);
     status = rm_power_management(sp, nv, 0, NV_PM_ACPI_STANDBY);
 
+    nvidia_pci_save_config_space(sp, dev);
+
     NV_UMA_ZONE_FREE_STACK(sp);
 
     return (status == RM_OK) ? 0 : EIO;
@@ -206,6 +209,8 @@ int nvidia_resume(device_t dev)
     if (sp == NULL)
         return ENOMEM;
 
+    nvidia_pci_restore_config_space(sp, dev);
+
     sc = device_get_softc(dev);
     nv = sc->nv_state;
 
@@ -367,18 +372,19 @@ int nvidia_get_card_info(void *args, int size)
             continue;
         nv = sc->nv_state;
 
-        ci[i].flags          = NV_IOCTL_CARD_INFO_FLAG_PRESENT;
-        ci[i].domain         = nv->domain;
-        ci[i].bus            = nv->bus;
-        ci[i].slot           = nv->slot;
-        ci[i].vendor_id      = nv->vendor_id;
-        ci[i].device_id      = nv->device_id;
-        ci[i].gpu_id         = nv->gpu_id;
-        ci[i].interrupt_line = nv->interrupt_line;
-        ci[i].fb_address     = nv->fb->address;
-        ci[i].fb_size        = nv->fb->size;
-        ci[i].reg_address    = nv->regs->address;
-        ci[i].reg_size       = nv->regs->size;
+        ci[i].flags              = NV_IOCTL_CARD_INFO_FLAG_PRESENT;
+        ci[i].pci_info.domain    = nv->pci_info.domain;
+        ci[i].pci_info.bus       = nv->pci_info.bus;
+        ci[i].pci_info.slot      = nv->pci_info.slot;
+        ci[i].pci_info.vendor_id = nv->pci_info.vendor_id;
+        ci[i].pci_info.device_id = nv->pci_info.device_id;
+        ci[i].gpu_id             = nv->gpu_id;
+        ci[i].interrupt_line     = nv->interrupt_line;
+        ci[i].fb_address         = nv->fb->cpu_address;
+        ci[i].fb_size            = nv->fb->size;
+        ci[i].reg_address        = nv->regs->cpu_address;
+        ci[i].reg_size           = nv->regs->size;
+        ci[i].minor_number       = i;
     }
 
     return 0;
@@ -535,7 +541,7 @@ int nvidia_open_dev(
             if (!sc->UD_sg_list)
                 goto failed;
 
-            sglist_append_phys(sc->UD_sg_list, nv->ud.address, nv->ud.size);
+            sglist_append_phys(sc->UD_sg_list, nv->ud.cpu_address, nv->ud.size);
 
             sc->UD_object = NV_VM_PAGER_ALLOCATE(OBJT_SG, sc->UD_sg_list,
                     nv->ud.size, (VM_PROT_READ | VM_PROT_WRITE),
@@ -543,9 +549,9 @@ int nvidia_open_dev(
             if (!sc->UD_object)
                 goto failed;
 
-            VM_OBJECT_LOCK(sc->UD_object);
+            VM_OBJECT_WLOCK(sc->UD_object);
             vm_object_set_memattr(sc->UD_object, VM_MEMATTR_UNCACHEABLE);
-            VM_OBJECT_UNLOCK(sc->UD_object);
+            VM_OBJECT_WUNLOCK(sc->UD_object);
         }
 
         nv->flags |= NV_FLAG_OPEN;
@@ -786,42 +792,21 @@ NvS32 NV_API_CALL nv_get_event(
 
 void* NV_API_CALL nv_alloc_kernel_mapping(
     nv_state_t *nv,
-    NvU64       address,
+    void       *pAllocPrivate,
+    NvU64       pageIndex,
+    NvU32       pageOffset,
     NvU64       size,
-    void      **private
+    void      **ppPrivate
 )
 {
-    struct nvidia_alloc *at;
-    struct nvidia_softc *sc = nv->os_state;
-    uint32_t i;
-    vm_offset_t offset, virtual_address;
+    struct nvidia_alloc *at = pAllocPrivate;
+    vm_offset_t virtual_address;
     int status;
 
-    offset = (vm_offset_t)address & PAGE_MASK;
-    address &= ~PAGE_MASK;
-
-    SLIST_FOREACH(at, &sc->alloc_list, list) {
-        for (i = 0; i < (at->size / PAGE_SIZE); i++) {
-            if (!at->alloc_type_contiguous &&
-                    (address == at->pte_array[i].physical_address))
-                goto found;
-            else if (at->alloc_type_contiguous &&
-                     (address == ((i * PAGE_SIZE) +
-                          at->pte_array[0].physical_address)))
-                goto found;
-        }
-    }
-
-    return NULL;
-
-found:
-    if ((size + (i * PAGE_SIZE)) > at->size)
-        return NULL;
-
     if (at->alloc_type_contiguous) {
-        *private = NULL;
-        return (void *)(at->pte_array[0].virtual_address +
-                    (i * PAGE_SIZE) + offset);
+        *ppPrivate = NULL;
+        return (void *)(NvUPtr)(at->pte_array[0].virtual_address +
+                (pageIndex * PAGE_SIZE) + pageOffset);
     }
 
     size = (size + PAGE_MASK) & ~PAGE_MASK;
@@ -829,8 +814,8 @@ found:
     vm_object_reference(at->object);
     virtual_address = vm_map_min(kernel_map);
 
-    status = vm_map_find(kernel_map, at->object, (i * PAGE_SIZE),
-            &virtual_address, size, VMFS_ANY_SPACE,
+    status = vm_map_find(kernel_map, at->object, (pageIndex * PAGE_SIZE),
+            &virtual_address, size, 0, VMFS_ANY_SPACE,
             (VM_PROT_READ | VM_PROT_WRITE),
             (VM_PROT_READ | VM_PROT_WRITE), 0);
     if (status != KERN_SUCCESS) {
@@ -847,23 +832,24 @@ found:
         return NULL;
     }
 
-    *private = (void *)(NvUPtr)size;
-    return (void *)(virtual_address + offset);
+    *ppPrivate = (void *)(NvUPtr)size;
+
+    return (void *)(virtual_address + pageOffset);
 }
 
 RM_STATUS NV_API_CALL nv_free_kernel_mapping(
     nv_state_t *nv,
+    void       *pAllocPrivate,
     void       *address,
-    void       *private,
-    NvBool      guest_mapping
+    void       *pPrivate
 )
 {
     vm_offset_t virtual_address;
     uint32_t size;
 
-    if (private != NULL) {
+    if (pPrivate != NULL) {
         virtual_address = (vm_offset_t)address & ~PAGE_MASK;
-        size = (NvUPtr)private;
+        size = (NvUPtr)pPrivate;
         vm_map_remove(kernel_map, virtual_address,
             (virtual_address + size));
     }
@@ -871,23 +857,43 @@ RM_STATUS NV_API_CALL nv_free_kernel_mapping(
     return RM_OK;
 }
 
-RM_STATUS NV_API_CALL nv_alloc_user_mapping(
+RM_STATUS NV_API_CALL nv_create_user_mapping_context(
     nv_state_t *nv,
-    NvU64       address,
-    NvU64       size,
-    NvU32       protect,
-    NvU64      *user_address,
-    void      **private
+    NvU64 address,
+    NvU64 size,
+    void **ppPrivate
 )
 {
-    *user_address = address;
+    *ppPrivate = NULL;
+    return RM_OK;
+}
+
+RM_STATUS NV_API_CALL nv_alloc_user_mapping(
+    nv_state_t *nv,
+    void       *pAllocPrivate,
+    NvU64       pageIndex,
+    NvU32       pageOffset,
+    NvU64       size,
+    NvU32       protect,
+    NvU64      *pUserAddress,
+    void      **ppPrivate
+)
+{
+    struct nvidia_alloc *at = pAllocPrivate;
+
+    if (at->alloc_type_contiguous)
+        *pUserAddress = (at->pte_array[0].physical_address + (pageIndex * PAGE_SIZE) + pageOffset);
+    else
+        *pUserAddress = (at->pte_array[pageIndex].physical_address + pageOffset);
+
     return RM_OK;
 }
 
 RM_STATUS NV_API_CALL nv_free_user_mapping(
     nv_state_t *nv,
-    NvU64       user_address,
-    void       *priv_data
+    void       *pAllocPrivate,
+    NvU64       userAddress,
+    void       *pPrivate
 )
 {
     return RM_OK;
@@ -945,7 +951,7 @@ NvS32 nv_alloc_contig_pages(
         return ENOMEM;
     }
 
-    address = kmem_alloc_contig(kernel_map, size, flags, 0,
+    address = kmem_alloc_contig(kmem_arena, size, flags, 0,
             sc->dma_mask, PAGE_SIZE, 0, attr);
     if (!address) {
         status = ENOMEM;
@@ -975,9 +981,9 @@ NvS32 nv_alloc_contig_pages(
         goto failed;
     }
 
-    VM_OBJECT_LOCK(at->object);
+    VM_OBJECT_WLOCK(at->object);
     vm_object_set_memattr(at->object, attr);
-    VM_OBJECT_UNLOCK(at->object);
+    VM_OBJECT_WUNLOCK(at->object);
 
     *private = at;
     SLIST_INSERT_HEAD(&sc->alloc_list, at, list);
@@ -994,8 +1000,7 @@ failed:
         os_flush_cpu_cache();
 
     if (at->pte_array[0].virtual_address != NULL) {
-        kmem_free(kernel_map,
-                at->pte_array[0].virtual_address, at->size);
+        kmem_free(kmem_arena, at->pte_array[0].virtual_address, at->size);
         malloc_type_freed(M_NVIDIA, at->size);
     }
 
@@ -1021,8 +1026,7 @@ NvS32 nv_free_contig_pages(
     if (at->attr != VM_MEMATTR_WRITE_BACK)
         os_flush_cpu_cache();
 
-    kmem_free(kernel_map, at->pte_array[0].virtual_address,
-            at->size);
+    kmem_free(kmem_arena, at->pte_array[0].virtual_address, at->size);
     malloc_type_freed(M_NVIDIA, at->size);
 
     free(at->pte_array, M_NVIDIA);
@@ -1085,7 +1089,7 @@ NvS32 nv_alloc_system_pages(
     }
 
     for (i = 0; i < count; i++) {
-        address = kmem_alloc_contig(kernel_map, PAGE_SIZE, flags, 0,
+        address = kmem_alloc_contig(kmem_arena, PAGE_SIZE, flags, 0,
                 sc->dma_mask, PAGE_SIZE, 0, attr);
         if (!address) {
             status = ENOMEM;
@@ -1118,9 +1122,9 @@ NvS32 nv_alloc_system_pages(
         goto failed;
     }
 
-    VM_OBJECT_LOCK(at->object);
+    VM_OBJECT_WLOCK(at->object);
     vm_object_set_memattr(at->object, attr);
-    VM_OBJECT_UNLOCK(at->object);
+    VM_OBJECT_WUNLOCK(at->object);
 
     *private = at;
     SLIST_INSERT_HEAD(&sc->alloc_list, at, list);
@@ -1139,8 +1143,7 @@ failed:
     for (i = 0; i < count; i++) {
         if (at->pte_array[i].virtual_address == 0)
             break;
-        kmem_free(kernel_map,
-                at->pte_array[i].virtual_address, PAGE_SIZE);
+        kmem_free(kmem_arena, at->pte_array[i].virtual_address, PAGE_SIZE);
         malloc_type_freed(M_NVIDIA, PAGE_SIZE);
     }
 
@@ -1169,8 +1172,7 @@ NvS32 nv_free_system_pages(
         os_flush_cpu_cache();
 
     for (i = 0; i < count; i++) {
-        kmem_free(kernel_map,
-                at->pte_array[i].virtual_address, PAGE_SIZE);
+        kmem_free(kmem_arena, at->pte_array[i].virtual_address, PAGE_SIZE);
         malloc_type_freed(M_NVIDIA, PAGE_SIZE);
     }
 
@@ -1193,17 +1195,6 @@ RM_STATUS NV_API_CALL nv_alias_pages(
     return RM_ERR_NOT_SUPPORTED;
 }
 
-RM_STATUS NV_API_CALL nv_guest_pfn_list(
-    nv_state_t  *nv,
-    unsigned int key,
-    unsigned int pfn_count,
-    unsigned int offset_index,
-    unsigned int *user_pfn_list
-)
-{
-    return RM_ERR_NOT_SUPPORTED;
-}
-
 RM_STATUS NV_API_CALL nv_alloc_pages(
     nv_state_t *nv,
     NvU32       count,
@@ -1220,12 +1211,12 @@ RM_STATUS NV_API_CALL nv_alloc_pages(
     if (!alloc_type_contiguous) {
         if (nv_alloc_system_pages(nv, count, cache_type, zero,
                     pte_array, private)) {
-            status = RM_ERR_NO_FREE_MEM;
+            status = RM_ERR_NO_MEMORY;
         }
     } else {
         if (nv_alloc_contig_pages(nv, count, cache_type, zero,
                     pte_array, private)) {
-            status = RM_ERR_NO_FREE_MEM;
+            status = RM_ERR_NO_MEMORY;
         }
     }
 
@@ -1284,8 +1275,9 @@ NvU64 NV_API_CALL nv_get_user_phys_address(NvU64 address)
     return pmap_extract(vmspace_pmap(vm), va);
 }
 
-int nvidia_mmap_dev_single(
+int nvidia_mmap_ctl_single(
     nv_state_t *nv,
+    struct nvidia_filep *filep,
     vm_ooffset_t *offset,
     vm_size_t size,
     vm_object_t *object
@@ -1295,57 +1287,71 @@ int nvidia_mmap_dev_single(
     nv_stack_t *sp;
     RM_STATUS rmStatus;
     struct nvidia_softc *sc = nv->os_state;
-    NvU32 i, prot;
+    NvU64 pageIndex;
+    NvU32 prot;
+
+    sp = sc->api_sp;
+
+    rmStatus = rm_acquire_api_lock(sp);
+    if (rmStatus != RM_OK)
+        return EAGAIN;
+
+    rmStatus = rm_validate_mmap_request(sp, nv, filep, *offset,
+            size, &prot, (void **)&at, &pageIndex);
+    if (rmStatus != RM_OK) {
+        rm_release_api_lock(sp);
+        return EINVAL;
+    }
+
+    vm_object_reference(at->object);
+    *object = at->object;
+    *offset = (pageIndex * PAGE_SIZE);
+
+    rm_release_api_lock(sp);
+
+    return 0;
+}
+
+int nvidia_mmap_dev_single(
+    nv_state_t *nv,
+    struct nvidia_filep *filep,
+    vm_ooffset_t *offset,
+    vm_size_t size,
+    vm_object_t *object
+)
+{
+    nv_stack_t *sp;
+    RM_STATUS rmStatus;
+    struct nvidia_softc *sc = nv->os_state;
+    NvU32 prot;
 
     sp = sc->api_sp;
 
     NV_PCI_CHECK_CONFIG_SPACE(sp, nv, TRUE, TRUE, TRUE);
 
-    rmStatus = rm_validate_mmap_request(sp, nv, *offset, size, &prot);
+    rmStatus = rm_validate_mmap_request(sp, nv, filep, *offset, size,
+            &prot, NULL, NULL);
     if (rmStatus != RM_OK)
         return EINVAL;
 
     if (IS_UD_OFFSET(nv, *offset, size)) {
         *object = sc->UD_object;
         vm_object_reference(*object);
-        *offset = (*offset - nv->ud.address);
+        *offset = (*offset - nv->ud.cpu_address);
         return 0;
     } else if (IS_FB_OFFSET(nv, *offset, size)) {
         *object = sc->BAR_objects[NV_GPU_BAR_INDEX_FB];
         vm_object_reference(*object);
-        *offset = (*offset - nv->fb->address);
+        *offset = (*offset - nv->fb->cpu_address);
         return 0;
     } else if (IS_REG_OFFSET(nv, *offset, size)) {
         *object = sc->BAR_objects[NV_GPU_BAR_INDEX_REGS];
         vm_object_reference(*object);
-        *offset = (*offset - nv->regs->address);
+        *offset = (*offset - nv->regs->cpu_address);
         return 0;
     }
 
-    SLIST_FOREACH(at, &sc->alloc_list, list) {
-        for (i = 0; i < (at->size / PAGE_SIZE); i++) {
-            if (!at->alloc_type_contiguous &&
-                    (*offset == at->pte_array[i].physical_address))
-                goto found;
-            else if (at->alloc_type_contiguous &&
-                     (*offset == ((i * PAGE_SIZE) +
-                          at->pte_array[0].physical_address)))
-                goto found;
-        }
-    }
-
     return EINVAL;
-
-found:
-    if ((size + (i * PAGE_SIZE)) > at->size)
-        return EINVAL;
-
-    vm_object_reference(at->object);
-
-    *object = at->object;
-    *offset = (i * PAGE_SIZE);
-
-    return 0;
 }
 
 void nvidia_rc_timer(void *data)
@@ -1421,8 +1427,8 @@ void* NV_API_CALL nv_get_adapter_state(
             continue;
         nv = sc->nv_state;
 
-        if ((nv->domain == domain) &&
-                (nv->bus == bus) && (nv->slot == slot)) {
+        if ((nv->pci_info.domain == domain) &&
+            (nv->pci_info.bus == bus) && (nv->pci_info.slot == slot)) {
             return (void *) nv;
         }
     }
@@ -1452,22 +1458,21 @@ void* NV_API_CALL nv_get_smu_state(void)
     return NULL;
 }
 
-RM_STATUS NV_API_CALL nv_lock_user_pages(
+RM_STATUS NV_API_CALL nv_dma_map_pages(
     nv_state_t *nv,
-    void       *address,
-    NvU64       count,
+    NvU64       page_count,
     NvU64      *pte_array,
-    void      **private
+    void      **priv
 )
 {
     return RM_ERR_NOT_SUPPORTED;
 }
 
-RM_STATUS NV_API_CALL nv_unlock_user_pages(
+RM_STATUS NV_API_CALL nv_dma_unmap_pages(
     nv_state_t *nv,
-    NvU64       count,
+    NvU64       page_count,
     NvU64      *pte_array,
-    void       *private
+    void      **priv
 )
 {
     return RM_ERR_NOT_SUPPORTED;
@@ -1493,3 +1498,23 @@ RM_STATUS NV_API_CALL nv_alloc_os_descriptor_handle(
 {
     return RM_ERR_NOT_SUPPORTED;
 }
+
+NvU64 NV_API_CALL nv_get_dma_start_address(
+    nv_state_t *nv
+)
+{
+    return 0;
+}
+
+void NV_API_CALL nv_pci_trigger_recovery(
+    nv_state_t *nv
+)
+{
+}
+
+RM_STATUS NV_API_CALL nv_set_primary_vga_status(
+    nv_state_t *nv
+)
+{
+    return RM_ERR_NOT_SUPPORTED;
+}
